/*
 * stellarisMS.c
 *
 *  Created on: 11.08.2008
 *      Author: Christian Loeschen
 */
#include <curl/curl.h>
#include <curl/easy.h>
#include <raptor.h>
#include <stdarg.h>
#include "rsApiHandler.h"
#include "modAVUMetadata.h"
#include "reGlobalsExtern.h"
#include "stellarisMS.h"

#define HEADER_LENGTH 128		// KANN EVTL WEG?
#define STR_LEN 1024

char * createRdf(char *subject, char *predicate, char *object, char *rdfRepresentation);
int sendToStellaris(char *subject, const char *rdf_graph, msParam_t *outParam);
size_t curlWriter(void *ptr, size_t size, size_t nmemb, void *stream);
int addMetadata(char *file, char *att, char *val, char *unit, ruleExecInfo_t *rei);
int readProperties();

char *return_message;
struct properties {
	char *stellaris_url;
	char *x509_cert_file;
	char *x509_key_file;
	char *cacert_file;
	char *url_graph_prefix;
	char *proxy;
} *props;

/**
 * Test function
 * @param[in,out]	rei		the rule execution information
 * @return			the status code, 0 on success
 */
int
msiStellarisTest(ruleExecInfo_t *rei) {

	RE_TEST_MACRO("    msiStellarisTest test log")
	rodsLog(LOG_NOTICE, "this is stellaris test log");
	return 0;
}

/**
 * Test function for printing data.
 * @param[in]		content	the content to print
 * @param[in,out]	rei		the rule execution information
 * @return			the status code, 0 on success
 */
int
msiStellarisPrint(msParam_t *content, ruleExecInfo_t* rei) {

	RE_TEST_MACRO("    msiStellarisPrint test log")
	rodsLog(LOG_NOTICE, "log: %s", content->inOutStruct);
	return 0;
}

/**
 * Adds the parameters (AVU) to the given file.
 * @param[in]		msParam_t *file			the file to store the AVU-triple on
 * @param[in]		msParam_t *att			the attribute of the AVU-triple
 * @param[in]		msParam_t *val			the value of the AVU-triple
 * @param[in]		msParam_t *unit			the unit of the AVU-triple
 * @param[in,out]	rei						the rule execution information
 * @return			the status code, 0 on success
 */
int
msiStellarisAddMetadata(msParam_t *file, msParam_t *att, msParam_t *val, msParam_t *unit,
        ruleExecInfo_t *rei) {

	RE_TEST_MACRO("    msiStellarisAddMetadata test log")

	//	rodsLog(LOG_NOTICE, "rei->status = %i, %s", rei->status, rei->statusStr);

	rei->status = addMetadata(file->inOutStruct, att->inOutStruct, val->inOutStruct, unit->inOutStruct, rei);
	return rei->status;
}

int
addMetadata(char *file, char *att, char *val, char *unit, ruleExecInfo_t *rei) {

	modAVUMetadataInp_t modAVUMetadataInp;
	modAVUMetadataInp.arg0 = "add";
	modAVUMetadataInp.arg1 = "-d";
	modAVUMetadataInp.arg2 = file;
	modAVUMetadataInp.arg3 = att;
	modAVUMetadataInp.arg4 = val;
	modAVUMetadataInp.arg5 = unit;

	return rsModAVUMetadata(rei->rsComm, &modAVUMetadataInp);
}

/**
 * Creates a rdf representation from the given values, based on rdfRepresentation.
 * @param[in]		char *subject		the rdf-subject
 * @param[in]		char *predicate		the rdf-predicate
 * @param[in]		char *object		the rdf-object
 * @param[in]		char *rdfRepresentation		the rdf representation (rdfxml, rdfxml-abbrev, ...)
 * @param[out]		char *outParam		the created rdf-grpah
 * @param[in,out]	rei					the rule execution information
 * @return			the status code, 0 on success
 */
int
msiStellarisCreateRdf(msParam_t *subject, msParam_t *predicate, msParam_t *object,
        msParam_t *rdfRepresentation, msParam_t *outParam, ruleExecInfo_t *rei) {

	RE_TEST_MACRO("    msiStellarisCreateRdf test log")

	char *rdf_graph = NULL;

	rdf_graph = createRdf(subject->inOutStruct, predicate->inOutStruct, object->inOutStruct,
	        rdfRepresentation->inOutStruct);

	if (NULL != rdf_graph) {

		fillStrInMsParam(outParam, rdf_graph);
		rei->status = 0; //ok
	}
	else {
		rei->status = 1; //error
	}
	//	rei->status = 4;
	//	sprintf(rei->statusStr, "test");
	return rei->status;
}

char *
createRdf(char *subject, char *predicate, char *object, char *rdfRepresentation) {

	char *rdf_statement = NULL;

	readProperties();
	raptor_init();

	raptor_serializer *serializer = NULL;
	raptor_statement *triple;

	// make a triple with URI subject, URI predicate, literal object
	triple = (raptor_statement *) malloc(sizeof(raptor_statement));

	if (NULL != props->url_graph_prefix) {

		rdf_statement = (char *) malloc(sizeof(char) * (strlen(props->url_graph_prefix) + strlen(subject)));
		strncpy(rdf_statement, props->url_graph_prefix, strlen(props->url_graph_prefix));
		strncat(rdf_statement, subject, strlen(subject));
		triple->subject = (void *) raptor_new_uri((const unsigned char *) rdf_statement);
		triple->subject_type = RAPTOR_IDENTIFIER_TYPE_RESOURCE;
		free(rdf_statement);
		rdf_statement = NULL;
	}
	else {
		triple->subject = (void *) raptor_new_uri((const unsigned char *) subject);
	}
	triple->subject_type = RAPTOR_IDENTIFIER_TYPE_RESOURCE;

	triple->predicate = (void *) raptor_new_uri((const unsigned char *) predicate);
	triple->predicate_type = RAPTOR_IDENTIFIER_TYPE_RESOURCE;

	triple->object = object;
	triple->object_type = RAPTOR_IDENTIFIER_TYPE_LITERAL;
	//	triple->object_literal_language = (const unsigned char *) "en";		//	?

	serializer = raptor_new_serializer(rdfRepresentation);

	if (serializer) {
		/*
		 * no specific namespace used
		 *
		 * 	// create namespace
		 const unsigned char *ns_prefix = (const unsigned char *) "stellaris";
		 raptor_uri *ns_uri = raptor_new_uri(
		 (unsigned char *) "http://stellaris.zib.de/schema/atomic_operations#");

		 if (raptor_serialize_set_namespace(serializer, ns_uri, ns_prefix)) {

		 rodsLog(LOG_ERROR, "raptor_serialize_set_namespace failed");
		 return NULL;
		 }
		 */
		// start serializing
		if (raptor_serialize_start_to_string(serializer, NULL, (void **) &rdf_statement, NULL)) {

			rodsLog(LOG_ERROR, "raptor_serialize_start_to_string failed");
			return NULL;
		}
		// add the triple
		if (raptor_serialize_statement(serializer, triple)) {

			rodsLog(LOG_ERROR, "raptor_serialize_statement failed");
			return NULL;
		}
		// free the triple
		raptor_free_uri((raptor_uri *) triple->subject);
		raptor_free_uri((raptor_uri *) triple->predicate);
		free(triple);

		// finish serializing
		raptor_serialize_end(serializer);
		raptor_free_serializer(serializer);
	}
	// stop now
	raptor_finish();


	//rodsLog(LOG_NOTICE, "%s\n", rdf_statement);
	return rdf_statement;
}

/**
 * Creates a html request as stellaris wants it with the given values and sends it.
 * @param[in]		msParam_t *file			the file the content belongs to
 * @param[in]		msParam_t *content		the content to send
 * @param[out]		msParam_t *outParam		the created rdf-grpah
 * @param[in,out]	rei		the rule execution information
 * @return			the status code, 0 on success
 */
int
msiStellarisSendToStellaris(msParam_t *file, msParam_t *content, msParam_t *retCode, msParam_t *outParam,
        ruleExecInfo_t *rei) {

	RE_TEST_MACRO("    msiStellarisSendToStellaris test log")

	return_message = (char *) malloc(STR_LEN * sizeof(char));
	sprintf(return_message, "");

	rei->status = sendToStellaris(file->inOutStruct, content->inOutStruct, retCode);
	fillStrInMsParam(outParam, return_message);
	free(return_message);

	return rei->status;
}

int
sendToStellaris(char *subject, const char *rdf_graph, msParam_t *retCode) {

	char *content_header = "Content-Type: application/rdf+xml";
	char slug_header[HEADER_LENGTH] = "Slug: ";
	char errorBuffer[CURL_ERROR_SIZE];
	char *tmp_message;
	CURL *curlHandle;
	CURLcode res = CURLE_FAILED_INIT;
	long response_code, ssl_verify = 0;

	tmp_message = (char *) malloc(1024 * sizeof(char));
	res = curl_global_init(CURL_GLOBAL_DEFAULT);
	curlHandle = curl_easy_init();

	if (NULL == curlHandle) {

		rodsLog(LOG_ERROR, "curl_easy_init failed for rdf %s", rdf_graph);
	}
	else {
		readProperties();

		if (NULL == props->stellaris_url) {

			rodsLog(LOG_NOTICE, "missing stellaris_url property");
			return -1;
		}

		// create request
		struct curl_slist *headers = NULL;
		headers = curl_slist_append(headers, content_header);

		if (NULL == headers) {

			rodsLog(LOG_ERROR, "curl_slist_append failed on header %s", content_header);
		}
		if (strlen(subject) > 1) {

			strncat(slug_header, subject, strlen(subject));
			headers = curl_slist_append(headers, slug_header);

			if (NULL == headers) {

				rodsLog(LOG_ERROR, "curl_slist_append failed on header %s", slug_header);
			}
		}

		//	set all options
		if ((res = curl_easy_setopt(curlHandle, CURLOPT_ERRORBUFFER, errorBuffer)) != CURLE_OK) {

			rodsLog(LOG_ERROR, "curl_easy_setopt %s failed. CurlCode: %i", "CURLOPT_ERRORBUFFER", res);
		}
		if ((res = curl_easy_setopt(curlHandle, CURLOPT_HTTPHEADER, headers)) != CURLE_OK) {

			rodsLog(LOG_ERROR, "curl_easy_setopt %s failed. CurlCode: %i", "CURLOPT_HTTPHEADER", res);
		}
		if ((res = curl_easy_setopt(curlHandle, CURLOPT_URL, props->stellaris_url)) != CURLE_OK) {

			rodsLog(LOG_ERROR, "curl_easy_setopt %s failed. CurlCode: %i", "CURLOPT_URL", res);
		}
		if ((res = curl_easy_setopt(curlHandle, CURLOPT_POSTFIELDS, rdf_graph)) != CURLE_OK) {

			rodsLog(LOG_ERROR, "curl_easy_setopt %s failed. CurlCode: %i", "CURLOPT_POSTFIELDS", res);
		}
		if ((res = curl_easy_setopt(curlHandle, CURLOPT_WRITEFUNCTION, *curlWriter)) != CURLE_OK) {

			rodsLog(LOG_ERROR, "curl_easy_setopt %s failed. CurlCode: %i", "CURLOPT_WRITEFUNCTION", res);
		}
		if (NULL != props->proxy) {

			if ((res = curl_easy_setopt(curlHandle, CURLOPT_PROXY, props->proxy)) != CURLE_OK) {

				rodsLog(LOG_ERROR, "curl_easy_setopt %s failed. CurlCode: %i", "CURLOPT_PROXY", res);
			}
		}
		if (NULL != props->x509_cert_file) {

			if ((res = curl_easy_setopt(curlHandle, CURLOPT_SSLCERT, props->x509_cert_file)) != CURLE_OK) {

				rodsLog(LOG_ERROR, "curl_easy_setopt %s failed. CurlCode: %i", "CURLOPT_SSLCERT", res);
			}
		}
		if (NULL != props->x509_key_file) {

			if ((res = curl_easy_setopt(curlHandle, CURLOPT_SSLKEY, props->x509_key_file)) != CURLE_OK) {

				rodsLog(LOG_ERROR, "curl_easy_setopt %s failed. CurlCode: %i", "CURLOPT_SSLKEY", res);
			}
		}
		if (NULL != props->cacert_file) {

			if ((res = curl_easy_setopt(curlHandle, CURLOPT_CAINFO, props->cacert_file)) != CURLE_OK) {

				rodsLog(LOG_ERROR, "curl_easy_setopt %s failed. CurlCode: %i", "CURLOPT_CAINFO", res);
			}
			ssl_verify = 1;
		}
		if ((res = curl_easy_setopt(curlHandle, CURLOPT_SSL_VERIFYPEER, ssl_verify)) != CURLE_OK) {

			rodsLog(LOG_ERROR, "curl_easy_setopt %s failed. CurlCode: %i", "CURLOPT_SSL_VERIFYPEER", res);
		}
		if ((res = curl_easy_setopt(curlHandle, CURLOPT_SSL_VERIFYHOST, ssl_verify)) != CURLE_OK) {

			rodsLog(LOG_ERROR, "curl_easy_setopt %s failed. CurlCode: %i", "CURLOPT_SSL_VERIFYHOST", res);
		}

		//	perform request
		if ((res = curl_easy_perform(curlHandle)) != CURLE_OK) {

			rodsLog(LOG_ERROR, "curl_easy_perform failed. CurlCode: %i, Error: %s", res, errorBuffer);
		}
		if ((res = curl_easy_getinfo(curlHandle, CURLINFO_RESPONSE_CODE, &response_code)) != CURLE_OK) {

			rodsLog(LOG_ERROR, "curl_easy_getinfo %s failed. CurlCode: %i, Error: %s",
			        "CURLINFO_RESPONSE_CODE", res, errorBuffer);
		}

		if (res) {

			sprintf(tmp_message, "error: cannot receive response code\n%s", return_message);
		}
		else {
			if (strlen(return_message) > 0) {

				sprintf(tmp_message, "response code: %ld\nresponse message: %s", response_code,
				        return_message);
			}
			else {
				sprintf(tmp_message, "response code: %ld\n", response_code);
			}
			char *tmp = (char *) malloc(10 * sizeof(char));
			sprintf(tmp, "%ld", response_code);
			fillStrInMsParam(retCode, tmp);
		}
		strncpy(return_message, tmp_message, strlen(tmp_message));

		curl_easy_cleanup(curlHandle);
	}
	curl_global_cleanup();

	return res;
}

/**
 * Writes the stellaris answer into return_message.
 */
size_t
curlWriter(void *ptr, size_t size, size_t nmemb, void *stream) {

	char *item_new = return_message;
	char *item = strstr((char *) ptr, "<body>");
	char *end = strstr((char *) ptr, "</body>");
	short tag = 0;
	int pos = 0;

	if (NULL == item || NULL == end) {

		sprintf(return_message, "%s", (char *) ptr);
	}
	else { // parse text from html message
		item += 6;

		for (; item != end && pos < STR_LEN; item = item + 1) {

			if (*item == '<') {

				tag = 0;
			}
			else if (*item == '>') {

				tag = 1;
			}
			else if (tag) {

				if (!(*item == ' ' && (*(item - 1) == ' ' || *(item + 1) == ' '))) {

					if (*item == '\n' && *(item_new - 2) == '\n') {

						item_new -= 2;
						pos -= 2;
					}
					else {
						*item_new++ = *item;
						pos++;
					}
				}
			}
		}
		*item_new = '\0';
	}
	//rodsLog(LOG_NOTICE, "stellaris answer:\n%s", return_message);
	return nmemb;
}

/**
 * Read properties from stellaris.properties file.
 * stellaris.properties in <iRODS path>/server/config/
 */
int
readProperties() {

	FILE *file;
	char property[512];
	char *value, *tmp;
	char *dir = malloc(sizeof(char) * (strlen(getConfigDir()) + 25 /*strlen("/mod_stellaris.properties")*/));

	props = (struct properties *) malloc(sizeof(struct properties));
	props->stellaris_url = NULL;
	props->x509_cert_file = NULL;
	props->x509_key_file = NULL;
	props->cacert_file = NULL;
	props->url_graph_prefix = NULL;
	props->proxy = NULL;

	sprintf(dir, "%s/mod_stellaris.properties", getConfigDir());
	file = fopen(dir, "r");

	if (NULL == file) {

		rodsLog(LOG_ERROR, "cannot read properties from %s", dir);
		return 1;
	}

	while ((fscanf(file, "%[^\n]\n", property)) != EOF) {

		if ('#' != property[0]) {

			value = strchr(property, ':');

			if (NULL != value) {

				*value++ = '\0';
				tmp = (char *) malloc(sizeof(char) * strlen(value));
				strcpy(tmp, value);

				if (strcmp(property, "stellaris-url") == 0) {

					props->stellaris_url = tmp;
				}
				else if (strcmp(property, "x509-cert-file") == 0) {

					props->x509_cert_file = tmp;
				}
				else if (strcmp(property, "x509-key-file") == 0) {

					props->x509_key_file = tmp;
				}
				else if (strcmp(property, "cacert-file") == 0) {

					props->cacert_file = tmp;
				}
				else if (strcmp(property, "url_graph_prefix") == 0) {

					props->url_graph_prefix = tmp;
				}
				else if (strcmp(property, "proxy") == 0) {

					props->proxy = tmp;
				}
				else {
					free(tmp);
				}
				//rodsLog(LOG_NOTICE, "property: %s, value: %s", property, ptr);
			}
		}
	}
	fclose(file);
	return 0;
}
